/*  admiral.go - Thread per Client Admiral of the M.A.R-T>Y. Fleet
    Copyright (C) 2023  Matthew Hoffman

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>. */

package main

import (
	"fmt"
	"net"
	"os"
	"strconv"
	"sync"
	"time"

	"hoff.industries/marty/libmarty"
	"hoff.industries/marty/libmmp"
)

var TARGET_PROTOCOL = []int{2, 0, 0}
var INSTANCE_VERSION = []int{3, 0, 0}

const INSTANCE_NAME string = "m-goadmiral"
const INSTANCE_TITLE string = "Admiral - Commander of the Fleet"

const (
	CLIENT_STATE_UNDEFINED = 0
	CLIENT_STATE_PREINIT = 1
	CLIENT_STATE_ACTIVE = 2
	CLIENT_STATE_PINGING = 3
	CLIENT_STATE_TERMINATED = 4
)

type tracker struct {

	txIndex int
	rxIndex int

	txClient *libmarty.MartyClientParameters
	rxClient *libmarty.MartyClientParameters

	lock sync.Mutex
}

type clientConfig struct {
	id int
	state int

	aliveTries int
	lastAlive time.Time

	parameters libmarty.MartyClientParameters

	connection net.Conn
	netlock sync.Mutex

	rxQueue []*tracker
	txQueue []*tracker

	length int
	next *clientConfig
}

type clientContainer struct {

	idTable []*clientConfig
	masterMap map[libmarty.MartyClientParameters]*clientConfig

	hostMap map[string]*clientConfig
	nameMap map[string]*clientConfig
	prioMap map[uint16]*clientConfig

	lock sync.Mutex
}

func tracker_factory(messageTracker *tracker) {
	messageTracker.txIndex = 0
	messageTracker.rxIndex = 0

	messageTracker.txClient = nil
	messageTracker.rxClient = nil
}

func client_config_factory(config *clientConfig) {

	config.id = 0
	config.state = CLIENT_STATE_UNDEFINED

	config.aliveTries = 0
	config.lastAlive = time.Now()

	libmarty.Marty_client_parameters_factory(&config.parameters)

	config.connection = nil

	config.rxQueue = make([]*tracker, libmarty.MAX_MESSAGE_ID)
	config.txQueue = make([]*tracker, libmarty.MAX_MESSAGE_ID)

	for i := 0; i < libmarty.MAX_MESSAGE_ID; i++ {

		config.rxQueue[i] = nil
		config.txQueue[i] = nil
	}

	config.length = 0
	config.next = nil
}

func client_container_factory(container *clientContainer) {

	container.idTable = make([]*clientConfig, libmarty.MAX_CLIENT_ID)

	for i := 0; i < len(container.idTable); i++ {
		container.idTable[i] = nil
	}

	container.masterMap = make(map[libmarty.MartyClientParameters]*clientConfig)

	container.hostMap = make(map[string]*clientConfig)
	container.nameMap = make(map[string]*clientConfig)
	container.prioMap = make(map[uint16]*clientConfig)
}

func config_ll_add(head *clientConfig, add *clientConfig) {

	var current *clientConfig = head

	if current == nil {
		return
	}

	for current.next != nil {

		fmt.Println(current.next.parameters)
		current = current.next
	}

	current.next = add
	head.length++
}

func config_ll_rem(head *clientConfig, rem *clientConfig) {

	var current *clientConfig = head

	if current == nil {
		return
	}

	for current.next != rem {

		current = current.next

		if current == nil {
			return
		}
	}

	current.next = rem.next
	head.length--
}

func get_new_message_id(queue []*tracker) int{

	for i := 0; i < len(queue); i++ {

		if (queue[i] == nil) {
			return i
		}
	}
	return -1
}

func get_new_client_id(clientTable []*clientConfig) int{

	for i := 0; i < len(clientTable); i++ {

		if clientTable[i] == nil {
			return i
		}
	}

	return -1
}

func blocking_read(socketPointer *net.Conn, data chan []byte) {

	var networkError error
	var bytesRead int
	var buffer []byte

	var socket net.Conn = *socketPointer

	for true {

		buffer = make([]byte, libmmp.MAX_SIZE)

		bytesRead, networkError = socket.Read(buffer)

		if bytesRead <= 0 || networkError != nil {

			data <- nil
			break
		}

		data <- buffer
	}

}

func client_loop(me *clientConfig,
		data *libmarty.MartyServerData, container *clientContainer) {

	var start int = 0

	var networkError error

	var status libmmp.MMStatus

	var txPacket libmmp.MMPacket
	var rxPacket libmmp.MMPacket

	var targetParameters libmarty.MartyClientParameters
	var targetConfig *clientConfig
	//var linkedConfig *clientConfig

	var messageTracker *tracker

	var integerConv int

	var mapLookup bool
	var dataChannel = make(chan []byte, 1)

	var sonarTimeout = libmarty.SONAR_INTERVAL

	//var overflowBuffer []byte

	libmarty.Marty_client_parameters_factory(&targetParameters)
	libmmp.Mmpacket_factory(&rxPacket, true)
	libmmp.Mmpacket_factory(&txPacket, true)

	go blocking_read(&me.connection, dataChannel)

	for (me.state != CLIENT_STATE_TERMINATED) {

		libmmp.Mmpacket_reset(&txPacket, true)
		targetConfig = nil
		//linkedConfig = nil

		if me.aliveTries > libmarty.SONAR_RETRIES {

			me.state = CLIENT_STATE_TERMINATED
			continue
		}

		if (start >= libmmp.MAX_SIZE - 1 || rxPacket.Bytes[start] == libmmp.DATA_EMPTY) {

			start = 0
			libmmp.Mmpacket_reset(&rxPacket, true)

			select {
				case <-time.After(time.Second * time.Duration(sonarTimeout)):

					if me.state != CLIENT_STATE_ACTIVE {

						continue
					}

					targetConfig = me
					txPacket.Type = libmmp.MMP_TYPE_SONAR
					txPacket.Argument = libmmp.ARGS_EMPTY

					me.aliveTries++

					if me.aliveTries > 1 {

						libmarty.Marty_server_client_log("Pinging client try %d of %d...\n",
							libmarty.LOG_WARNING, data, me.id,
								&me.parameters, me.aliveTries, libmarty.SONAR_RETRIES)
					}

					sonarTimeout = libmarty.SONAR_FAST_INTERVAL

					goto JMP_TX_PACKET

				case rxPacket.Bytes = <- dataChannel:
			}

			if rxPacket.Bytes == nil {

				libmarty.Marty_server_client_log("Client terminated connection.\n",
					libmarty.LOG_WARNING, data, me.id, &me.parameters)

				me.state = CLIENT_STATE_TERMINATED
				break
			}

			me.lastAlive = time.Now()
			me.aliveTries = 0

			sonarTimeout = libmarty.SONAR_INTERVAL
		}

		libmmp.Deserialize(&rxPacket, &status, start)
		if (status.Error == libmmp.UNTERMINATED_PACKET) {

			libmarty.Marty_server_client_log("pack too big overfglow! Library error 0x%x\n",
				libmarty.LOG_ERROR, data, me.id, &me.parameters, status.Error)
			continue
		}


		if (status.Error != libmmp.NONE) {

			libmarty.Marty_server_client_log("Couldn't deserialize packet! Library error 0x%x\n",
				libmarty.LOG_ERROR, data, me.id, &me.parameters, status.Error)

			for i := start; i < libmmp.MAX_SIZE; i++ {
				start++
				if (rxPacket.Bytes[i] == libmmp.TERM_PACKET) {
					break
				}
			}

			continue
		}

		start += status.Size

		switch (rxPacket.Type) {

			case libmmp.MMP_TYPE_SONAR:
				continue

			case libmmp.MMP_TYPE_INIT:

				targetConfig = me

				integerConv, networkError = strconv.Atoi(libmmp.Payload_2_str(rxPacket.Payloads[1]))

				if (networkError != nil) {

					libmarty.Marty_server_client_log("Couldn't convert priority! %s\n",
						libmarty.LOG_WARNING, data, me.id, &me.parameters, networkError)

					continue
				}

				me.parameters.Priority = uint16(integerConv)
				me.parameters.Host = libmmp.Payload_2_str(rxPacket.Payloads[2])
				me.parameters.Name = libmmp.Payload_2_str(rxPacket.Payloads[3])

				container.lock.Lock()

				_, mapLookup = container.masterMap[me.parameters]
				if mapLookup {

					libmarty.Marty_client_parameters_factory(&me.parameters)

					txPacket.Type = libmmp.MMP_TYPE_INVALID
					txPacket.Argument = libmmp.MMPArgs[txPacket.Type]["KEY_EXISTS"]
					txPacket.Payloads = [][]byte{
						{0x00},
					}

					container.lock.Unlock()

					libmarty.Marty_server_client_log("Key exists, init fail.\n",
						libmarty.LOG_ERROR, data, me.id, &me.parameters)

					goto JMP_TX_PACKET
				}

				/*
				linkedConfig, mapLookup = container.hostMap[me.parameters.Host]
				if !mapLookup {

					linkedConfig = new(clientConfig)
					client_config_factory(linkedConfig)
					container.hostMap[me.parameters.Host] = linkedConfig
				}

				config_ll_add(linkedConfig, me)

				linkedConfig, mapLookup = container.nameMap[me.parameters.Name]
				if !mapLookup {

					linkedConfig = new(clientConfig)
					client_config_factory(linkedConfig)
					container.nameMap[me.parameters.Name] = linkedConfig
				}

				config_ll_add(linkedConfig, me)

				linkedConfig, mapLookup = container.prioMap[me.parameters.Priority]
				if !mapLookup {

					linkedConfig = new(clientConfig)
					client_config_factory(linkedConfig)
					container.prioMap[me.parameters.Priority] = linkedConfig
				}

				config_ll_add(linkedConfig, me)*/

				container.masterMap[me.parameters] = me
				me.state = CLIENT_STATE_ACTIVE

				container.lock.Unlock()

				txPacket.Type = libmmp.MMP_TYPE_INIT
				txPacket.Argument = libmmp.MMPArgs[txPacket.Type]["ACCEPT"]
				txPacket.Payloads = rxPacket.Payloads

				libmarty.Marty_server_client_log("Initialisation complete.\n",
					libmarty.LOG_STARTUP, data, me.id, &me.parameters)

				goto JMP_TX_PACKET

			case libmmp.MMP_TYPE_CONTERM:
				me.state = CLIENT_STATE_TERMINATED
				continue

			case libmmp.MMP_TYPE_SEND:

				integerConv, networkError = strconv.Atoi(libmmp.Payload_2_str(rxPacket.Payloads[1]))

				if (networkError != nil) {

					libmarty.Marty_server_client_log("Couldn't decode target priority!.\n",
						libmarty.LOG_ERROR, data, me.id, &me.parameters)

					continue
				}

				targetParameters.Priority = uint16(integerConv)
				targetParameters.Host = libmmp.Payload_2_str(rxPacket.Payloads[2])
				targetParameters.Name = libmmp.Payload_2_str(rxPacket.Payloads[3])

				container.lock.Lock()
				targetConfig, mapLookup = container.masterMap[targetParameters]
				container.lock.Unlock()

				if !mapLookup {

					libmarty.Marty_server_client_log("Couldn't find target!\n",
						libmarty.LOG_ERROR, data, me.id, &me.parameters)
					continue
				}

				integerConv, networkError = strconv.Atoi(libmmp.Payload_2_str(rxPacket.Payloads[0]))

				if networkError != nil ||
						integerConv < 0 || me.txQueue[integerConv] != nil {

					libmarty.Marty_server_client_log("Message ID %d unavailable.\n",
						libmarty.LOG_ERROR, data, me.id, &me.parameters, integerConv)

					continue
				}

				messageTracker = new(tracker)
				tracker_factory(messageTracker)

				messageTracker.rxIndex = get_new_message_id(targetConfig.rxQueue)

				if messageTracker.rxIndex == -1 {

					libmarty.Marty_server_client_log("No message IDs available!\n",
						libmarty.LOG_ERROR, data, me.id, &me.parameters)
					continue
				}

				messageTracker.txIndex = integerConv
				messageTracker.txClient = &me.parameters

				targetConfig.rxQueue[messageTracker.rxIndex] = messageTracker

				rxPacket.Payloads[0] = []byte(fmt.Sprintf("%d", messageTracker.rxIndex))
				rxPacket.Payloads[1] = []byte(fmt.Sprintf("%d", me.parameters.Priority))
				rxPacket.Payloads[2] = []byte(me.parameters.Host)
				rxPacket.Payloads[3] = []byte(me.parameters.Name)

				txPacket.Type = libmmp.MMP_TYPE_SEND
				txPacket.Argument = libmmp.MMPArgs[libmmp.MMP_TYPE_SEND]["DIRECT"]
				txPacket.Payloads = rxPacket.Payloads

				goto JMP_TX_PACKET

			case libmmp.MMP_TYPE_APPLICATION:

				integerConv, networkError = strconv.Atoi(libmmp.Payload_2_str(rxPacket.Payloads[0]))

				if networkError != nil ||
					integerConv < 0 || me.rxQueue[integerConv] == nil {

						libmarty.Marty_server_client_log("Message ID %d was not valid.\n",
							libmarty.LOG_ERROR, data, me.id, &me.parameters, integerConv)
					continue
				}

				messageTracker = me.rxQueue[integerConv]

				if messageTracker == nil || messageTracker.txClient == nil {

					libmarty.Marty_server_client_log("Message tracker for ID %d does not exist.\n",
						libmarty.LOG_ERROR, data, me.id, &me.parameters, integerConv)
					continue
				}

				targetConfig, mapLookup = container.masterMap[*messageTracker.txClient]
				if !mapLookup {

					libmarty.Marty_server_client_log("Tracker target not found.\n",
						libmarty.LOG_ERROR, data, me.id, &me.parameters, integerConv)
					continue
				}

				txPacket.Type = libmmp.MMP_TYPE_APPLICATION
				txPacket.Argument = rxPacket.Argument
				txPacket.Payloads = [][]byte{
					[]byte(fmt.Sprintf("%d", messageTracker.txIndex)),
				}

				me.rxQueue[messageTracker.rxIndex] = nil
				targetConfig.txQueue[messageTracker.txIndex] = nil

				goto JMP_TX_PACKET


			case libmmp.MMP_TYPE_QUERY:
			default:

				libmarty.Marty_server_client_log("Got an unknown message type: 0x%x\n",
					libmarty.LOG_WARNING, data, me.id, &me.parameters, rxPacket.Type)
				continue
		}

		JMP_TX_PACKET:

		if targetConfig == nil {

			libmarty.Marty_server_client_log("No target! Won't send.\n",
				libmarty.LOG_ERROR, data, me.id, &me.parameters)
			continue
		}

		libmmp.Serialize(&txPacket, &status)
		if status.Error != libmmp.NONE {

			libmarty.Marty_server_client_log("Couldn't serialize packet! Library error 0x%x\n",
				libmarty.LOG_ERROR, data, me.id, &me.parameters, status.Error)
			continue
		}

		targetConfig.netlock.Lock()

		targetConfig.connection.Write(txPacket.Bytes[0:status.Size])

		targetConfig.netlock.Unlock()

		if txPacket.Type == libmmp.MMP_TYPE_SONAR {

			continue
		}

		libmarty.Marty_server_client_log("Type 0x%x Args 0x%x --> %s(%d | %s | %s)%s\n",
			libmarty.LOG_INFO, data, me.id, &me.parameters, txPacket.Type, txPacket.Argument,
				libmarty.ClientColours[targetConfig.id % len(libmarty.ClientColours)],
					targetConfig.parameters.Priority, targetConfig.parameters.Host, targetConfig.parameters.Name,
						libmarty.ClearColour)

	}

	libmarty.Marty_server_client_log("Terminating client and exiting thread.\n",
		libmarty.LOG_CRITICAL, data, me.id, &me.parameters)

	me.connection.Close()

	container.lock.Lock()

	/*
	linkedConfig, mapLookup = container.hostMap[me.parameters.Host]
	if mapLookup {
		config_ll_rem(linkedConfig, me)
	}

	linkedConfig, mapLookup = container.nameMap[me.parameters.Name]
	if mapLookup {
		config_ll_rem(linkedConfig, me)
	}

	linkedConfig, mapLookup = container.prioMap[me.parameters.Priority]
	if mapLookup {
		config_ll_rem(linkedConfig, me)
	}*/

	container.idTable[me.id] = nil
	delete(container.masterMap, me.parameters)
	container.lock.Unlock()
}

func socket_listener(listenType string, addr string,
		connectionChannel chan net.Conn, data *libmarty.MartyServerData) {

	var listener net.Listener
	var networkError error
	var connection net.Conn

	if (listenType == "unix") {
		os.Remove(data.UnixSocket)
	}

	listener, networkError = net.Listen(listenType, addr)
	if networkError != nil || listener == nil {

		libmarty.Marty_server_log("Error spawning %s listener! %s\n", libmarty.LOG_CRITICAL,
			data, listenType, networkError)

		os.Exit(libmarty.EXIT_ERR_SPAWN)
	}

	for true {
		connection, networkError = listener.Accept()

		if networkError != nil {

			continue
		}

		connectionChannel <- connection

		libmarty.Marty_server_log("Accepted connection of type %s\n", libmarty.LOG_INFO,
			data, listenType)
	}
}

func main() {
	var data libmarty.MartyServerData
	var container clientContainer

	var client *clientConfig

	var hostname string = ""
	var callError error = nil

	var connectionChannel chan net.Conn = make(chan net.Conn)

	libmarty.Marty_server_data_factory(&data)
	client_container_factory(&container)

	//data.TcpPort = libmarty.TestPort
	//data.UnixSocket = libmarty.TestSocket

	data.VersionString = fmt.Sprintf("%s v%d.%d.%d", INSTANCE_NAME,
		INSTANCE_VERSION[0], INSTANCE_VERSION[1], INSTANCE_VERSION[2])

	fmt.Printf("%s[%s]%s Starting %s %s - %s\n",
		libmarty.LoggerColours[libmarty.LOG_INFO], libmarty.MartyFlavour,
	libmarty.ClearColour, libmarty.LIBMARTY_AUTHOR, INSTANCE_TITLE, data.VersionString)

	hostname, callError = os.Hostname()
	if (callError != nil) {

		os.Exit(libmarty.EXIT_ERR_HOSTNAME)
	}

	data.MyParameters.Host = hostname
	data.MyParameters.Name = "admiral"

	go socket_listener("tcp", data.TcpHost + ":" + data.TcpPort, connectionChannel, &data)
	libmarty.Marty_server_log("Started listener on TCP host %s port %s\n", libmarty.LOG_STARTUP,
		&data, data.TcpHost, data.TcpPort)

	go socket_listener("unix", data.UnixSocket, connectionChannel, &data)
	libmarty.Marty_server_log("Started listener on UNIX socket %s\n", libmarty.LOG_STARTUP,
		&data, data.UnixSocket)

	for true {

		client = new(clientConfig)
		client_config_factory(client)

		client.connection = <- connectionChannel

		container.lock.Lock()
		client.id = get_new_client_id(container.idTable)

		if client.id == -1 {

			container.lock.Unlock()
			libmarty.Marty_server_log("Client connections maxxed out! Rejecting connection.\n",
				libmarty.LOG_STARTUP, &data)
			continue
		}

		container.idTable[client.id] = client
		container.lock.Unlock()

		libmarty.Marty_server_log("Client connected with id %d.\n",
			libmarty.LOG_STARTUP, &data, client.id)
		go client_loop(client, &data, &container)
	}

}
