/*	libmarty.go - M.A.R-T>Y. Go lib helper function definition file
	Copyright (C) 2023-2024 hoff.industries

	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <https://www.gnu.org/licenses/>. */

package libmarty

import (
	"fmt"
	"log"
	"net"
	"os"
	"strconv"
	"sync"
	"time"

	"hoff.industries/marty/libmmp"
)

const LIBMARTY_AUTHOR string = "hoff.industries"

var LIBMARTY_VERSION = []int{3, 1, 3}

const DefaultSocket string = "~/.config/marty/entrypoint"
const TestSocket string = "~/.config/marty/testpoint"

const DefaultHost string = "127.0.0.1"
const DefaultPort string = "9341"
const TestPort string = "9342"

const preinit string = "__preinit__"

const MartyColour string = "\033[38;5;45m"
const ClearColour string = "\033[0m"

const MartyFlavour string = "M.A.R-T>Y."
const BracketFlavour string = "[M.A.R-T>Y.]"

const GO_DAEMON_COLOUR int = 2
const RECONNECT_INTERVAL int = 2
const PAYL_IDX_ENDPOINT int = 4

const SONAR_RETRIES = 3
const SONAR_FAST_INTERVAL = 5
const SONAR_INTERVAL = 30

const (
	MAX_MESSAGE_ID int = 4096
	MAX_MESSAGE_ID_LEN = 5

	MAX_CLIENT_ID int = 4096
	MAX_CLIENT_ID_LEN int = 5
)

const (
	EXIT_ERR_CDATA int = 1
	EXIT_ERR_VSTRING int = 2
	EXIT_ERR_HOSTNAME int = 3

	EXIT_ERR_SPAWN int = 4
	EXIT_ERR_INIT int = 5

	EXIT_ERR_POLL int = 6
	EXIT_ERR_ITC int = 7

	EXIT_ERR_FORMAT int = 8
	EXIT_ERR_CHILD int = 9
	EXIT_ERR_SIGN int = 10
	EXIT_ERR_READ int = 11
	EXIT_ERR_BAD_POINTER int = 12
)

const (
	CONN_UNIX int = 0
	CONN_TCP int = 1
)

const (
	LOG_NOMINAL int = 0
	LOG_INFO int = 1
	LOG_WARNING int = 2
	LOG_ERROR int = 3
	LOG_CRITICAL int = 4
	LOG_STARTUP int = 5
	LOG_NOTITLE int = 6
	LOG_SUCCESS int = 7
)

var ClientColours = [9]string{
	"\033[38;5;48m", // spring green
	"\033[38;5;220m", // gold
	"\033[38;5;27m", // dark blue
	"\033[38;5;196m", // red
	"\033[38;5;201m", // magenta
	"\033[38;5;208m", // dark orange
	"\033[38;5;124m", // dark red
	"\033[38;5;93m", // purple
	"\033[38;5;34m", // green
}

var CONNECTION_TYPE = map[int]string{

	CONN_UNIX: "unix",
	CONN_TCP: "tcp",
}

var LoggerColours = map[int]string{

	LOG_NOMINAL: "\033[38;5;48m",
	LOG_INFO: MartyColour,
	LOG_WARNING: "\033[0;38;5;202m",
	LOG_ERROR: "\033[38;5;124m",
	LOG_CRITICAL: "\033[38;5;196m",
	LOG_STARTUP: "\033[38;5;93m",
	LOG_NOTITLE: "",
	LOG_SUCCESS: "\033[38;5;48m",
}

type MartyClientParameters struct {
	Auth string
	Priority uint16
	Host string
	Name string
}

type MartyMessageHandle struct {
	Result uint16
	Execute *libmmp.MMPacket
	Next *MartyMessageHandle
}

type MartyClientEndpoint struct {
	Endpoint int
	Name string

	Min int
	Max int

	IsAscii bool
}

type MartyServerData struct {
	VersionString string

	TcpHost string
	TcpPort string

	UnixSocket string

	MyParameters MartyClientParameters
	InfoLogger *log.Logger
}

type MartyClientData struct {
	VersionString string

	// connection settings
	TargetHost string
	TargetPort string
	TargetSocket string

	ConnType int
	LogLevel int

	// connection objects
	socket net.Conn
	networkLock sync.Mutex

	// network event channel and received message
	MessageChannel chan *libmmp.MMPacket
	DeliveryChannel chan *MartyMessageHandle

	InitParameters MartyClientParameters
	MyParameters MartyClientParameters

	deliveryTracker [MAX_MESSAGE_ID]*MartyMessageHandle
	deliveryLock sync.Mutex
}

func Marty_client_parameters_factory(parameters *MartyClientParameters) {
	parameters.Auth = preinit
	parameters.Priority = 0
	parameters.Host = preinit
	parameters.Name = preinit
}

func Marty_server_data_factory(data *MartyServerData) {
	data.VersionString = preinit

	data.TcpHost = DefaultHost
	data.TcpPort = DefaultPort
	data.UnixSocket = DefaultSocket

	data.InfoLogger = log.New(os.Stdout, "", 0)

	Marty_client_parameters_factory(&data.MyParameters)
}

func Marty_client_data_factory(data *MartyClientData) {
	data.VersionString = preinit

	data.TargetHost = DefaultHost
	data.TargetPort = DefaultPort
	data.TargetSocket = DefaultSocket

	data.ConnType = CONN_TCP
	data.LogLevel = LOG_NOMINAL

	data.MessageChannel = make(chan *libmmp.MMPacket, MAX_MESSAGE_ID)
	data.DeliveryChannel = make(chan *MartyMessageHandle, MAX_MESSAGE_ID)

	Marty_client_parameters_factory(&data.InitParameters)
	Marty_client_parameters_factory(&data.MyParameters)

	for i := 0; i < MAX_MESSAGE_ID; i++ {
		data.deliveryTracker[i] = nil
	}
}

func Marty_client_endpoint_factory(endpoint *MartyClientEndpoint) {
	endpoint.Endpoint = -1
	endpoint.Name = preinit
	endpoint.Min = -1
	endpoint.Max = -1
}

func Marty_message_handle_factory(handle *MartyMessageHandle) {
	handle.Result = libmmp.MMPArgs[libmmp.MMP_TYPE_UNSET]["UNSET"]
	handle.Execute = nil
	handle.Next = nil
}

func Marty_part_log(message string, level int) {
	fmt.Printf("%s[%s]%s %s", LoggerColours[level], MartyFlavour, ClearColour, message)
}

func Marty_full_log(message string, level int, data *MartyClientData, parameters ...any) {

	appendMessage := fmt.Sprintf(message, parameters...)

	fmt.Printf("%s[%s] %s<%s> %s(%d | %s | %s)%s %s " + appendMessage, LoggerColours[level],
		MartyFlavour, ClientColours[GO_DAEMON_COLOUR], data.VersionString, MartyColour,
			data.MyParameters.Priority, data.MyParameters.Host, data.MyParameters.Name,
		ClearColour, time.Now().Format(time.RFC3339))
}

func Marty_server_log(message string, level int, data *MartyServerData, parameters ...any) {

	appendMessage := fmt.Sprintf(message, parameters...)

	data.InfoLogger.Printf("%s[%s] %s<%s>%s %s " + appendMessage, LoggerColours[level],
		MartyFlavour, ClientColours[GO_DAEMON_COLOUR], data.VersionString,
		ClearColour, time.Now().Format(time.RFC3339))
}

func Marty_server_client_log(message string, level int, data *MartyServerData,
		id int, client *MartyClientParameters, parameters ...any) {

	prefix := fmt.Sprintf("%s(%d | %s | %s)%s ", ClientColours[id % len(ClientColours)],
		client.Priority, client.Host, client.Name, ClearColour)

	Marty_server_log(prefix + message, level, data, parameters...)
}

func Marty_set_application(packet *libmmp.MMPacket, status *libmmp.MMStatus,
	messageId []byte, argument uint16) {

	var payloadLen int = libmmp.Payloadlen(messageId)

	libmmp.Mmstatus_factory(status)

	if (len(messageId) > MAX_MESSAGE_ID_LEN - 1) {
		status.Error = libmmp.MESSAGE_ID_UNAVAILABLE
		return
	}

	packet.Type = libmmp.MMP_TYPE_APPLICATION
	packet.Argument = argument

	packet.Payloads = make([][]byte, libmmp.PayloadLengths[libmmp.MMP_TYPE_APPLICATION])
	packet.Payloads[0] = make([]byte, MAX_MESSAGE_ID_LEN)

	for i := 0; i < payloadLen; i++ {
		packet.Payloads[0][i] = messageId[i]
	}

	packet.Payloads[0][payloadLen] = libmmp.TERM_PAYLOAD
}

func Marty_net_write(txPacket *libmmp.MMPacket, status *libmmp.MMStatus, data *MartyClientData) {

	data.networkLock.Lock()
	data.socket.Write(txPacket.Bytes)
	data.networkLock.Unlock()
}

func Get_new_message_id(data *MartyClientData) int{
	var i int = 0

	for i = 0; i < MAX_MESSAGE_ID; i++ {
		if (data.deliveryTracker[i] == nil) {

			return i
		}
	}

	return -1
}

func Marty_message_handle_free(handle *MartyMessageHandle, clearList bool) {

	if (handle == nil) {
		return
	}

	handle.Result = libmmp.MMPArgs[libmmp.MMP_TYPE_UNSET]["UNSET"]
	if (handle.Execute != nil) {

		libmmp.Mmpacket_reset(handle.Execute, true)
		handle.Execute = nil
	}

	if (clearList) {

		Marty_message_handle_free(handle.Next, true)
	}

	// NOTE: does this update for the caller? no idea
	handle = nil
}

func Marty_message_handler(handle *MartyMessageHandle, status *libmmp.MMStatus, data *MartyClientData) {

	var messageId int = -1

	libmmp.Mmstatus_factory(status)

	status.Size = -1

	if (handle == nil) {
		return
	}

	if (handle.Execute == nil) {
		status.Error = libmmp.UNSET_PACKET
		return
	}

	if (handle.Execute.Type != libmmp.MMP_TYPE_SEND) {
		status.Error = libmmp.INVALID_TYPE
		return
	}

	data.deliveryLock.Lock()

	messageId = Get_new_message_id(data)

	if (messageId == -1) {

		status.Error = libmmp.MESSAGE_ID_UNAVAILABLE
		data.deliveryLock.Unlock()
		return
	}

	if (handle.Execute.Payloads == nil ||
			len(handle.Execute.Payloads) < libmmp.PayloadLengths[libmmp.MMP_TYPE_SEND]) {

		status.Error = libmmp.INVALID_PAYLOAD_COUNT
		data.deliveryLock.Unlock()
		return
	}

	handle.Execute.Payloads[0] = []byte(fmt.Sprintf("%d", messageId))
	libmmp.Serialize(handle.Execute, status)

	if (status.Error != libmmp.NONE) {
		return
	}

	data.deliveryTracker[messageId] = handle
	data.deliveryLock.Unlock()

	status.Size = messageId
	Marty_net_write(handle.Execute, status, data)
}

func Marty_init(data *MartyClientData) libmmp.MMPError{

	var txPacket libmmp.MMPacket
	var rxPacket libmmp.MMPacket
	var status libmmp.MMStatus

	var mmpErr libmmp.MMPError = libmmp.NONE

	var bytesRead int = -1
	var err error = nil

	libmmp.Mmpacket_factory(&txPacket, true)
	libmmp.Mmpacket_factory(&rxPacket, true)
	libmmp.Mmstatus_factory(&status)

	txPacket.Type = libmmp.MMP_TYPE_INIT
	txPacket.Argument = libmmp.MMPArgs[libmmp.MMP_TYPE_INIT]["INIT"]
	txPacket.Payloads = make([][]byte, libmmp.PayloadLengths[libmmp.MMP_TYPE_INIT])

	txPacket.Payloads[0] = make([]byte, len(data.InitParameters.Auth) + 1)
	txPacket.Payloads[1] = make([]byte, 2)
	txPacket.Payloads[2] = make([]byte, len(data.InitParameters.Host) + 1)
	txPacket.Payloads[3] = make([]byte, len(data.InitParameters.Name) + 1)

	txPacket.Payloads[1][0] = byte(data.InitParameters.Priority) + '0'
	txPacket.Payloads[1][1] = libmmp.TERM_PAYLOAD

	libmmp.Str_2_payload(data.InitParameters.Auth, txPacket.Payloads[0])
	libmmp.Str_2_payload(data.InitParameters.Host, txPacket.Payloads[2])
	libmmp.Str_2_payload(data.InitParameters.Name, txPacket.Payloads[3])

	libmmp.Serialize(&txPacket, &status)
	if (status.Error != libmmp.NONE) {

		mmpErr = status.Error
		goto JMP_CLEANUP_RETURN
	}

	Marty_net_write(&txPacket, &status, data)

	bytesRead, err = data.socket.Read(rxPacket.Bytes)
	if (err != nil || bytesRead <= 0) {

		mmpErr = libmmp.NETWORK
		goto JMP_CLEANUP_RETURN
	}


	libmmp.Deserialize(&rxPacket, &status, 0)
	if (status.Error != libmmp.NONE) {

		mmpErr = status.Error
		goto JMP_CLEANUP_RETURN
	}

	if (rxPacket.Type == libmmp.MMP_TYPE_INVALID &&
			rxPacket.Argument == libmmp.MMPArgs[libmmp.MMP_TYPE_INVALID]["AUTH"]) {

		mmpErr = libmmp.AUTH_FAIL
		goto JMP_CLEANUP_RETURN
	}

	if (rxPacket.Type != libmmp.MMP_TYPE_INIT) {

		mmpErr = libmmp.INVALID_TYPE
		goto JMP_CLEANUP_RETURN
	}

	if (rxPacket.Argument != libmmp.MMPArgs[libmmp.MMP_TYPE_INIT]["ACCEPT"]) {

		mmpErr = libmmp.INIT_FAIL
		goto JMP_CLEANUP_RETURN
	}

	data.MyParameters.Auth = data.InitParameters.Auth
	data.MyParameters.Priority = data.InitParameters.Priority
	data.MyParameters.Host = data.InitParameters.Host
	data.MyParameters.Name = data.InitParameters.Name

	JMP_CLEANUP_RETURN:

	libmmp.Mmpacket_reset(&txPacket, true)
	libmmp.Mmpacket_reset(&rxPacket, true)

	return mmpErr
}

func GetServerLog(instanceName string, instanceVersion, printString string) string{
	return fmt.Sprintf("(%s v%s) %s: %s", instanceName,
		instanceVersion, time.Now().Format(time.RFC3339), printString)
}

func Spawn_marty_socket(data *MartyClientData) libmmp.MMPError{

	var unixAddr = net.UnixAddr{
		Name: data.TargetSocket,
		Net: "unix",
	}

	var networkError error = nil

	switch (data.ConnType) {
		case CONN_TCP:
			data.socket, networkError = net.Dial("tcp", data.TargetHost+":"+data.TargetPort)

		case CONN_UNIX:
			data.socket, networkError = net.DialUnix("unix", nil, &unixAddr)

		default:
			return libmmp.INVALID_TYPE
	}

	if networkError != nil {
		return libmmp.NETWORK
	}

	return libmmp.NONE
}

func handle_reconnect(data *MartyClientData, tryInterval int) {

	var err libmmp.MMPError = libmmp.NETWORK

	var tries int = 0
	var tpm int = 1

	if (RECONNECT_INTERVAL < 60) {
		tpm = 60 / RECONNECT_INTERVAL
	}

	for (err != libmmp.NONE) {

		if (tries == tpm) {
			Marty_full_log("Multiple attempts failed. Trying more...\n", LOG_ERROR, data)
			tries = 0
		}

		err = Spawn_marty_socket(data)

		if (err == libmmp.NONE) {
			err = Marty_init(data)
		}

		if (err == libmmp.NONE) {
			return
		}

		tries++

		time.Sleep(time.Duration(tryInterval) * time.Second)
	}
}

func Marty_network_loop(data *MartyClientData) {

	var bytesRead int = 0
	var start int = 0

	var idConversion int = -1

	var overflowBuffer [libmmp.MAX_SIZE]byte

	var rxPacket libmmp.MMPacket
	var txPacket libmmp.MMPacket

	var channelPacket *libmmp.MMPacket = nil

	var status libmmp.MMStatus

	var handle *MartyMessageHandle = nil

	var networkError error = nil

	for i := 0; i < len(overflowBuffer); i++ {
		overflowBuffer[i] = 0x00
	}

	libmmp.Mmpacket_factory(&rxPacket, true)

	for true {

		libmmp.Mmpacket_factory(&txPacket, true)
		networkError = nil

		if (start >= libmmp.MAX_SIZE - 1 || rxPacket.Bytes[start] == libmmp.DATA_EMPTY) {

			start = 0

			libmmp.Mmpacket_reset(&rxPacket, true)
			bytesRead, networkError = data.socket.Read(rxPacket.Bytes)

			if (bytesRead <= 0 || networkError != nil) {
				Marty_full_log("Connection closed, reconnecting...\n", LOG_WARNING, data)

				handle_reconnect(data, RECONNECT_INTERVAL)
				Marty_full_log("Reconnection success!\n", LOG_STARTUP, data)

				//data.MessageChannel <- nil
				continue
			}
		}

		libmmp.Deserialize(&rxPacket, &status, start)
		if (status.Error != libmmp.NONE) {
			Marty_full_log("Couldn't deserialize packet! Library error 0x%x\n", LOG_ERROR, data, status.Error)

			for i := start; i < libmmp.MAX_SIZE; i++ {
				start++
				if (rxPacket.Bytes[i] == libmmp.TERM_PACKET) {
					break
				}
			}

			continue
		}

		start += status.Size

		switch (rxPacket.Type) {
			case libmmp.MMP_TYPE_SONAR:

				txPacket.Type = libmmp.MMP_TYPE_SONAR
				break

			case libmmp.MMP_TYPE_SEND:

				Marty_full_log("Direct message received.\n", LOG_INFO, data)
				channelPacket = libmmp.Packet_2_heap(&rxPacket, false, start - status.Size)
				if (channelPacket == nil) {
					Marty_full_log("Error allocating message to heap!\n", LOG_ERROR, data)
					continue
				}

				data.MessageChannel <- channelPacket
				continue

			case libmmp.MMP_TYPE_APPLICATION:

				Marty_full_log("Application message received.\n", LOG_INFO, data)

				idConversion, networkError = strconv.Atoi(libmmp.Payload_2_str(rxPacket.Payloads[0]))

				if (networkError != nil || idConversion < 0 || idConversion > MAX_MESSAGE_ID) {
					Marty_full_log("Error parsing application message id %d!\n", LOG_ERROR, data, idConversion)
					continue
				}

				data.deliveryLock.Lock()
				handle = data.deliveryTracker[idConversion]

				if (handle == nil) {

					Marty_full_log("Error mapping application message id!\n", LOG_ERROR, data)
					data.deliveryLock.Unlock()
					continue
				}

				data.deliveryTracker[idConversion] = nil
				data.deliveryLock.Unlock()

				handle.Result = rxPacket.Argument
				data.DeliveryChannel <- handle

				continue

			default:
				continue
		}

		libmmp.Serialize(&txPacket, &status)
		if (status.Error != libmmp.NONE) {
			Marty_full_log("Couldn't serialize autopacket!\n", LOG_ERROR, data)
			continue
		}

		Marty_net_write(&txPacket, &status, data)
	}
}
