/*	libmmp.go - M.A.R-T>Y. Go lib protocol implementation definition file
	Copyright (C) 2023-2024 hoff.industries

	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <https://www.gnu.org/licenses/>. */

package libmmp

//import "fmt"

var MMP_LIB_VERSION =			[]int{0, 1, 1}
var MMP_IMPLEMENTS =			[]int{2, 0, 0}

const LIBMMP_AUTHOR =			"hoff.industries"

const MIN_SIZE =				4
const MAX_SIZE =				4096
const MAX_CLIENTS =				20

const DATA_EMPTY uint8 =		0x00
const ARGS_EMPTY uint16 =		0x0000
const TERM_PAYLOAD uint8 =		0xFE
const TERM_PACKET uint8 =		0xFF

const PKT_IDX_PAYLOAD_START =	3

const PAYLOAD_FIELD_OPT =		0
const PAYLOAD_FIELD_OPT_R =		1
const PAYLOAD_FIELD_REQ =		2
const PAYLOAD_FIELD_NULL =		3

type ClientState int64
type MMPType uint8
type MessageDirection int64
type PayloadConfiguration int64

type MMPError int64

// define MMP message types
const (
	MMP_TYPE_UNSET MMPType =		iota
	MMP_TYPE_INIT MMPType =			0x01
	MMP_TYPE_INVALID MMPType =		0x02
	MMP_TYPE_CONTERM MMPType =		0x03
	MMP_TYPE_SONAR MMPType =		0x04
	MMP_TYPE_SEND MMPType =			0x05
	MMP_TYPE_APPLICATION MMPType =	0x06
	MMP_TYPE_QUERY MMPType =		0x07
)

const (
	NONE MMPError =					0
	NETWORK MMPError =				1
	MEMORY MMPError =				3
	FORMAT MMPError =				4

	INIT_FAIL MMPError =			5
	DELIVERY_FAIL MMPError =		6
	QUERY_FAIL MMPError =			7

	QUEUE_PUSH MMPError =			8
	QUEUE_PULL MMPError =			9

	UNSET_MESSAGE_TYPE =			10
	UNSET_ARGUMENT =				11
	UNSET_PAYLOADS =				12
	UNSET_PACKET =					13

	INVALID_TYPE =					14
	INVALID_ARGUMENT =				15
	INVALID_PACKET_LENGTH =			16
	INVALID_PAYLOAD_COUNT =			17
	INVALID_PAYLOAD_VALUE =			18
	INVALID_MAP_KEY =				19

	MESSAGE_ID_UNAVAILABLE =		20
	WAIT_INTERRUPTION =				21
	AUTH_FAIL =						22
	UNTERMINATED_PACKET = 			23
)

const (
	PREINIT ClientState =			0
	ACTIVE ClientState =			1
	INACTIVE ClientState =			2
)

const (
	OPT PayloadConfiguration =		0
	OPT_R PayloadConfiguration =	1
	REQ PayloadConfiguration =		2
)

type MMPacket struct {
	Type MMPType
	Argument uint16
	Payloads [][]byte

	Bytes []byte
}

type MMStatus struct {
	Size int
	Error MMPError
}

var MMPArgs = map[MMPType]map[string]uint16 {

	MMP_TYPE_UNSET:			{

		"UNSET":			0x0000,
	},

	MMP_TYPE_INIT:			{

		"INIT":				0x0001,
		"ACCEPT":			0x0001,
	},

	MMP_TYPE_INVALID:		{

		"TYPE":				0x0001,
		"LENGTH":			0x0002,
		"ARGS":				0x0003,
		"PAYLOAD":			0x0004,
		"PAYLOAD_COUNT":	0x0005,
		"AUTH":				0x0006,
		"KEY_EXISTS":		0x0007,
		"KEY_RESERVED":		0x0008,
		"MESSAGE_ID":		0x0009,
	},

	MMP_TYPE_CONTERM:		{

		"CLEAN":			0x0001,
		"SPAM":				0x0002,
		"AUTH":				0x0003,
	},

	MMP_TYPE_SEND:			{

		"DIRECT":			0x0001,
		"BROADCAST":		0x0002,
	},

	MMP_TYPE_APPLICATION:	{

		"SUCCESS":			0x0001,
		"DELIVERY":			0x0002,
		"BUSY":				0x0003,
		"RANGE":			0x0004,
		"ENDPOINT":			0x0005,
		"VALUE":			0x0006,
		"TOO_FEW":			0x0007,
		"TOO_MANY":			0x0008,
		"AUTH":				0x0009,
		"EXECUTE":			0x000A,
	},

	MMP_TYPE_QUERY:			{

		"CONN_NUMBER":		0x0001,
		"CLIENT_INFO":		0x0002,
		"NOT_FOUND":		0x0003,
	},

}

var PayloadLengths = map[MMPType]int {
	MMP_TYPE_UNSET:			0,
	MMP_TYPE_INIT:			4,
	MMP_TYPE_INVALID:		1,
	MMP_TYPE_CONTERM:		0,
	MMP_TYPE_SONAR:			0,
	MMP_TYPE_SEND:			5,
	MMP_TYPE_APPLICATION:	1,
	MMP_TYPE_QUERY:			3,
}

var ArgumentRange = map[MMPType]uint16 {

	MMP_TYPE_UNSET:			MMPArgs[MMP_TYPE_UNSET]["UNSET"],
	MMP_TYPE_INIT:			MMPArgs[MMP_TYPE_INIT]["ACCEPT"],
	MMP_TYPE_INVALID:		MMPArgs[MMP_TYPE_INVALID]["MESSAGE_ID"],
	MMP_TYPE_CONTERM:		MMPArgs[MMP_TYPE_CONTERM]["AUTH"],
	MMP_TYPE_SEND:			MMPArgs[MMP_TYPE_SEND]["BROADCAST"],
	MMP_TYPE_APPLICATION:	MMPArgs[MMP_TYPE_APPLICATION]["EXECUTE"],
	MMP_TYPE_QUERY:			MMPArgs[MMP_TYPE_QUERY]["NOT_FOUND"],
}

func Mmp_compatible(programVersion []int) bool{

	return (MMP_IMPLEMENTS[0] == programVersion[0] &&
		MMP_IMPLEMENTS[1] == programVersion[1] &&
			MMP_IMPLEMENTS[2] == programVersion[2])
}

func Mmpacket_factory(packet *MMPacket, zero bool) {
	packet.Type = MMP_TYPE_UNSET
	packet.Argument = MMPArgs[MMP_TYPE_UNSET]["UNSET"]
	packet.Payloads = nil

	if (packet.Bytes == nil) {
		packet.Bytes = make([]byte, MAX_SIZE)
	}

	if (zero) {
		for i := 0; i < len(packet.Bytes); i++ {
			packet.Bytes[i] = 0x00
		}
	}
}

func Stringable(payload []byte) bool{

	for i := 0; i < Payloadlen(payload); i++ {

		if (payload[i] < ' ' || payload[i] > '~') {
			return false
		}
	}

	return true
}

func Payloadlen(payload []byte) int{
	var i int = 0

	for i = 0; payload[i] != TERM_PAYLOAD; i++ {}

	return i
}

func Str_2_payload(input string, payload []byte) {
	var i int = 0

	for i = 0; i < len(input); i++ {
		payload[i] = input[i]
	}

	payload[i] = TERM_PAYLOAD
}

func Payload_2_str(payload []byte) string{
	var i int = 0
	var output string = ""

	for i = 0; i < Payloadlen(payload); i++ {

		if (payload[i] < ' ' || payload[i] > '~') {
			output = output + " "
		}

		output = output + string(payload[i])
	}

	return output
}

func Mmpacket_reset(packet *MMPacket, zero bool) {

	if (packet.Payloads == nil) {
		Mmpacket_factory(packet, zero)
		return
	}

	for i := 0; i < len(packet.Payloads); i++ {
		packet.Payloads[i] = nil
	}

	packet.Payloads = nil
	Mmpacket_factory(packet, zero)

}

func Mmstatus_factory(status *MMStatus) {
	status.Size = 0
	status.Error = NONE
}

func Serialize(packet *MMPacket, status *MMStatus) {

	var payloadLength int = -1
	var packetIndex int = PKT_IDX_PAYLOAD_START

	Mmstatus_factory(status)

	if (packet.Bytes == nil) {
		packet.Bytes = make([]byte, MAX_SIZE)
	}

	for i := 0; i < len(packet.Bytes); i++ {
		packet.Bytes[i] = 0x00
	}

	if (packet.Type == MMP_TYPE_UNSET) {
		status.Error = UNSET_MESSAGE_TYPE
		return
	}

	if (packet.Type > MMP_TYPE_QUERY) {
		status.Error = INVALID_TYPE
		return
	}

	if (packet.Argument > ArgumentRange[packet.Type]) {
		status.Error = INVALID_ARGUMENT
		return
	}

	payloadLength = PayloadLengths[packet.Type]
	if (payloadLength != 0 && packet.Payloads == nil) {
		status.Error = UNSET_PAYLOADS
		return
	}

	if (payloadLength == 0 && packet.Payloads != nil) {
		status.Error = INVALID_PAYLOAD_COUNT
		return
	}

	if (payloadLength > len(packet.Payloads)) {
		status.Error = INVALID_PAYLOAD_COUNT
		return
	}

	packet.Bytes[0] = byte(packet.Type)
	packet.Bytes[1] = byte(packet.Argument >> 8)
	packet.Bytes[2] = byte(packet.Argument & 0xFF)

	for i := 0; i < len(packet.Payloads); i++ {

		if (packet.Payloads[i] == nil) {

			status.Error = UNSET_PAYLOADS
			return
		}

		for j := 0; j < len(packet.Payloads[i]); j++ {

			if (packetIndex >= MAX_SIZE) {

				status.Error = INVALID_PACKET_LENGTH
				return
			}

			packet.Bytes[packetIndex] = packet.Payloads[i][j]
			packetIndex++

			if (packet.Payloads[i][j] == TERM_PAYLOAD) {
				break
			}

			if (j == len(packet.Payloads[i]) - 1 && packet.Payloads[i][j] != TERM_PAYLOAD) {
				packet.Bytes[packetIndex] = TERM_PAYLOAD
				packetIndex++
			}
		}
	}

	if (packetIndex >= MAX_SIZE) {
		status.Error = INVALID_PACKET_LENGTH
		return
	}

	packet.Bytes[packetIndex] = TERM_PACKET
	status.Size = packetIndex + 1
}

func Payloads_2_heap(payloads [][]byte) [][]byte{

	var newPayloads = make([][]byte, len(payloads))

	for i := 0; i < len(payloads); i++ {

		newPayloads[i] = make([]byte, len(payloads[i]))

		for j := 0; j < len(payloads[i]); j++ {
			newPayloads[i][j] = payloads[i][j]
		}
	}

	return newPayloads
}

func Deserialize(packet *MMPacket, status *MMStatus, start int) {

	var payloadCount int = 0
	var payloadIndex int = 0
	var payloadSize int = 0

	var packetIndex int = start + PKT_IDX_PAYLOAD_START

	var payloadSizes []int = nil

	Mmstatus_factory(status);
	Mmpacket_reset(packet, false)

	packet.Type = MMPType(packet.Bytes[start])
	if (packet.Type == MMP_TYPE_UNSET) {

		status.Error = INVALID_TYPE
		return
	}

	packet.Argument = uint16(packet.Bytes[start + 1]) << 8 | uint16(packet.Bytes[start + 2])

	if (packet.Argument > ArgumentRange[packet.Type]) {
		status.Error = INVALID_ARGUMENT
		return
	}

	for i := packetIndex; i < MAX_SIZE; i++ {

		if (packet.Bytes[i] == TERM_PAYLOAD) {
			payloadIndex++
		}

		if (packet.Bytes[i] == TERM_PACKET) {
			break
		}
	}

	payloadCount = PayloadLengths[packet.Type]

	if (payloadIndex < payloadCount) {
		status.Error = INVALID_PAYLOAD_COUNT
		return
	}

	if (payloadCount == 0) {
		status.Size = MIN_SIZE
		return
	}

	payloadSizes = make([]int, payloadIndex)
	payloadIndex = 0

	for i := packetIndex; i < MAX_SIZE; i++ {

		if (packet.Bytes[i] == TERM_PAYLOAD) {

			payloadSizes[payloadIndex] = payloadSize + 1
			payloadIndex++
			payloadSize = 0
			continue
		}

		if (packet.Bytes[i] == TERM_PACKET) {
			break
		}

		if (i == MAX_SIZE - 1) {

			status.Error = UNTERMINATED_PACKET
			return
		}

		payloadSize++

	}

	// TODO: Is this logic correct??? Am I checking the
	// right counter variable here?
	/*
	if (payloadCount < payloadIndex) {

		payloadSizes = nil
		status.Error = INVALID_PAYLOAD_COUNT
		return
	}*/

	packet.Payloads = make([][]byte, payloadIndex)

	payloadSize = PKT_IDX_PAYLOAD_START

	for i := 0; i < payloadIndex; i++ {

		packet.Payloads[i] = make([]byte, payloadSizes[i])

		for j := 0; j < payloadSizes[i]; j++ {

			packet.Payloads[i][j] = packet.Bytes[start + j + payloadSize]
		}
		payloadSize += payloadSizes[i]
	}

	status.Size = payloadSize + 1
	payloadSizes = nil
}

func Packet_2_heap(packet *MMPacket, doSerial bool, start int) *MMPacket{

	var heapPacket = new(MMPacket)
	var status MMStatus

	Mmstatus_factory(&status)
	Mmpacket_factory(heapPacket, true)

	if (doSerial) {
		Serialize(packet, &status)

		if (status.Error != NONE) {

			Mmpacket_reset(heapPacket, true)
			heapPacket = nil
			return nil
		}
	}

	for i := 0; i < MAX_SIZE; i++ {
		heapPacket.Bytes[i] = packet.Bytes[i]
	}

	Deserialize(heapPacket, &status, start)
	if (status.Error != NONE) {

		Mmpacket_reset(heapPacket, true)
		heapPacket = nil
	}

	return heapPacket
}
